#!/usr/bin/env python
import getopt, sys, os, subprocess

def usage():
    print """Lithium, an automated testcase reduction tool by Jesse Ruderman
    
Usage:
    
./lith.py [options] test file [eargs]

* test: an 'interestingness test' script
     It must return 0 for "interesting" and nonzero for "uninteresting".
* file: an 'interesting' file you would like reduced
     It must contain lines with "DDBEGIN" and "DDEND" marking the
     section of the file to be reduced.
* eargs: extra arguments to pass to the 'interestingness test' script

Global options:
* --strategy=[delta, remove-pair, remove-substring]. default: delta.
* --granularity=[lines, chars]. default: lines.

Additional options for the default strategy (--strategy=delta)
* --repeat=[always, last, never]. default: last
     Whether to repeat a chunk size if chunks are removed.
* --max=n. default: about half of the file.
* --min=n. default: 1.
     What chunk sizes to test.  Must be powers of two.
* --chunksize=n
     Shortcut for repeat=none, min=n, max=n

See readme.txt for more information.

"""


# Globals

strategy = "delta"
deltaRepeat = "last"
deltaMin = 1
deltaMax = pow(2, 30)
    
granularity = "lines"

testFilename = None
testArgs = None
testcaseFilename = None
testcaseExtension = ""

tempDir = None
tempFileCount = 0

before = ""
after = ""
parts = []
enabled = []
numParts = 0


# Main and friends

def main():
    global testFilename, testArgs, testcaseFilename, testcaseExtension, strategy

    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["help", "granularity=", "strategy=", "repeat=", "min=", "max=", "chunksize="])
    except getopt.GetoptError, exc:
        usageError(exc.msg)
        
    processOptions(opts)

    if len(args) < 2:
        usageError("You need to specify a test and a file.")
        
    testFilename = args[0]
    testcaseFilename = args[1]
    testArgs = args[:] # includes the name of the test program, which subprocess.call needs anyway

    e = testcaseFilename.rsplit(".", 1)
    if len(e) > 1:
        testcaseExtension = "." + e[1]


    readTestcase()
    
    print "The original testcase has " + str(numParts) + " " + granularity + "."
    
    print "Checking that the original testcase is 'interesting'..."
    if not interesting():
        usageError("The original testcase is not 'interesting'!")
        
    if numParts == 0:
        usageError("The file has 0 " + granularity + " so there's nothing for Lithium to try to remove!")

    strategyFunction = {
        'delta': delta,
        'remove-pair': tryRemovingPair,
        'remove-substring': tryRemovingSubstring
    }.get(strategy, None)

    if not strategyFunction:
        usageError("Unknown strategy!")

    createTempDir()
    print "Intermediate files will be stored in " + tempDir + "/."
    writeTestcaseTemp("original", False)

    strategyFunction()


def processOptions(opts):
    global granularity, deltaRepeat, deltaMin, deltaMax, strategy

    for o, a in opts:
        if o == "--granularity":
            granularity = a
            if not (granularity in ("chars", "lines")):
                usageError("granularity must be 'chars' or 'lines'.")
            
        if o == "--strategy":
            strategy = a
            
        if o == "--min":
            deltaMin = int(a)
            if not isPowerOfTwo(deltaMin):
                usageError("deltaMin must be a power of two.")
            
        if o == "--max":
            deltaMax = int(a)
            if not isPowerOfTwo(deltaMax):
                usageError("deltaMin must be a power of two.")
            
        if o == "--repeat":
            deltaRepeat = a
            if not (deltaRepeat in ("always", "last", "never")):
                usageError("deltaRepeat must be one of 'always', 'last', or 'never'.")
                
        if o == "--chunksize":
            deltaMin = int(a)
            deltaMax = deltaMin
            deltaRepeat = "never"
            if not isPowerOfTwo(deltaMin):
                usageError("deltaMin must be a power of two.")
            
        if o == "--help":
            usage()
            sys.exit()


def usageError(s):
    print s
    print "Use --help if you need it :)"
    sys.exit(2)



# Functions for manipulating the testcase (aka the 'interesting' file)

def readTestcase():
    global testcaseFilename
    global before, after
    global granularity
    global parts, numParts, enabled

    try:
        file = open(testcaseFilename, "r")
    except IOError:
        usageError("Can't read the original testcase file, " + testcaseFilename + "!")
    
    for line in file:
        before += line 
        if line.find("DDBEGIN") != -1:
            break
    else:
        usageError("The testcase (" + testcaseFilename + ") has no line containing the string 'DDBEGIN'.")

    for line in file:
        if line.find("DDEND") != -1:
            after += line
            break
        readTestcaseLine(line)
    else:
        usageError("The testcase (" + testcaseFilename + ") has no line containing the string 'DDEND'.")

    for line in file:
        after += line

    file.close()
    
    if granularity == "chars" and numParts > 0:
        # Move the line break at the end of the last line out of the reducible
        # part so the "DDEND" line doesn't get combined with another line.
        parts.pop()
        enabled.pop()
        numParts -= 1
        after = "\n" + after

    
def readTestcaseLine(line):
    global granularity
    global parts, numParts, enabled
    
    if granularity == "lines":
       parts.append(line)
       enabled.append(True)
       numParts += 1
    elif granularity == "chars":
        for char in line:
            parts.append(char)
            enabled.append(True)
            numParts += 1


def writeTestcase(filename):
    global numParts
    
    file = open(filename, "w")
    file.write(before)
    for i in range(numParts):
        if enabled[i]:
            file.write(parts[i])
    file.write(after)
    file.close()
   
   
def writeTestcaseTemp(partialFilename, useNumber):
    global tempFileCount
    if useNumber:
        tempFileCount += 1
        partialFilename = str(tempFileCount) + "-" + partialFilename
    writeTestcase(tempDir + "/" + partialFilename + testcaseExtension)


def createTempDir():
    global tempDir
    i = 1
    while 1:
        tempDir = "tmp" + str(i)
        if not os.path.exists(tempDir):
            os.mkdir(tempDir)
            break
        i += 1



# Interestingness test

def interesting():
    global tempFileCount, testcaseFilename, testArgs

    writeTestcase(testcaseFilename)

    try:
        status = subprocess.call(testArgs)
    except OSError, e:
        print "Lithium tried to run:"
        print "  " + repr(testArgs)
        print "but got this error:"
        print "  " + str(e)
        sys.exit(2)

    # Save an extra copy of the file inside the temp directory.
    # This is useful if you're reducing an assertion and encounter a crash:
    # it gives you a way to try to reproduce the crash.
    if tempDir != None:
        if status == 0:
          tempFileTag = "interesting"
        else:
          tempFileTag = "boring"
        writeTestcaseTemp(tempFileTag, True)

    return status == 0


# Main reduction algorithm
# With most parameters, guarantees that the result is 1-minimal.

def delta():

    chunkSize = min(deltaMax, largestPowerOfTwoSmallerThan(numParts))
    finalChunkSize = max(deltaMin, 1)
    
    while 1:
        anyChunksRemoved = tryRemovingChunks(chunkSize);
    
        last = (chunkSize == finalChunkSize)

        if anyChunksRemoved and (deltaRepeat == "always" or (deltaRepeat == "last" and last)):
            # Repeat with the same chunk size
            continue
        elif last:
            # Done
            writeTestcase(testcaseFilename)
            break
        else:
            # Continue with the next smaller chunk size
            chunkSize /= 2
            assert chunkSize >= 1
            continue


def tryRemovingChunks(chunkSize):
    """Make a single run through the testcase, trying to remove chunks of size chunkSize.
    
    Returns True iff any chunks were removed."""
    
    global enabled
    
    chunksSoFar = 0
    summary = ""

    chunksRemoved = 0
    chunksSurviving = 0
    granRemoved = 0
    granSurviving = 0

    print "Starting a round with chunks of " + str(chunkSize) + " " + granularity + "."

    for chunkStart in range(0, numParts, chunkSize):

        if not enabled[chunkStart]:
            continue

        chunkEnd = min(numParts, chunkStart + chunkSize)
        
        for pos in range(chunkStart, chunkEnd):
            enabled[pos] = False
        
        if interesting():
            print "Yay, reduced it by removing " + str(chunkStart) + ".." + str(chunkEnd - 1) + " :)"
            
            chunksRemoved += 1
            granRemoved += (chunkEnd - chunkStart)
            summary += '-';
        else:
            print "Removing " + str(chunkStart) + ".." + str(chunkEnd - 1) + " made the file 'uninteresting'.  Putting it back."

            chunksSurviving += 1
            granSurviving += (chunkEnd - chunkStart)
            summary += 'S';
    
            for pos in range(chunkStart, chunkEnd):
                enabled[pos] = True
  
        # Put a space between each pair of chunks in the summary.
        # During 'delta', this is useful because it shows visually which 
        # chunks used to be part of a single larger chunk.
        
        chunksSoFar += 1
        if chunksSoFar % 2 == 0:
          summary += " ";
  
    print ""
    print "Done with a round of chunk size " + str(chunkSize) + "!"
    print str(chunksSurviving) + " chunks survived; " + str(chunksRemoved) + " were removed."
    print str(granSurviving) + " " + granularity + " survived; " + str(granRemoved) + " were removed."
    print "Which chunks survived: " + summary
    print ""
    
    writeTestcaseTemp("did-round-" + str(chunkSize), True);
   
    return (chunksRemoved > 0)
    
    

# Other reduction algorithms
# (Use these if you're really frustrated with something you know is 1-minimal.)

def tryRemovingPair():
    global numParts

    for i in range(0, numParts):
        enabled[i] = False
        for j in range(i + 1, numParts):
            enabled[j] = False
            print "Trying removing the pair " + str(i) + ", " + str(j)
            if interesting():
                print "Success!  Removed a pair!  Exiting."
                sys.exit(0)
            enabled[j] = True
        enabled[i] = True
        
    # Restore the original testcase
    writeTestcase(testcaseFilename)
    print "Failure!  No pair can be removed."
            

def tryRemovingSubstring():
    global numParts

    for i in range(0, numParts):
        for j in range(i, numParts):
            enabled[j] = False
            print "Trying removing the substring " + str(i) + ".." + str(j)
            if interesting():
                print "Success!  Removed a substring!  Exiting."
                sys.exit(0)
        for j in range(i, numParts):
            enabled[j] = True

    # Restore the original testcase
    writeTestcase(testcaseFilename)
    print "Failure!  No substring can be removed."



# Math helpers

def isPowerOfTwo(n):
    i = 1
    while 1:
        if i == n:
            return True
        if i > n:
            return False
        i *= 2
    
def largestPowerOfTwoSmallerThan(n):
    i = 1
    while 1:
        if i * 2 >= n:
            return i
        i *= 2


# Run main

if __name__ == "__main__":
    main()
